/* This file is part of Mailfromd.             -*- c -*-
   Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#define DEFAULT_DB_MODE 0640
#include <fnmatch.h>
#include "mf-dbm.h"

struct db_prop {                /* Database properties */
	char *pat;              /* Database name pattern */
	mode_t mode;            /* File mode */
	int null;               /* Null byte */
};

static mu_list_t db_prop_list;

static int
db_prop_compare(const void *item, const void *data)
{
	const struct db_prop *prop = item;
	const char *name = data;
	return strcmp(prop->pat, name);
}

static int
strtomode(char *str, mode_t *pmode)
{
	mode_t mode = 0;
	struct { char c; unsigned mask; } modetab[] = {
		{ 'r', S_IRUSR },
		{ 'w', S_IWUSR },
		{ 'x', S_IXUSR },
		{ 'r', S_IRGRP },
		{ 'w', S_IWGRP },
		{ 'x', S_IXGRP },
		{ 'r', S_IROTH },
		{ 'w', S_IWOTH },
		{ 'x', S_IXOTH },
		{ 0 }
	};
	int i;

	for (i = 0; modetab[i].c; i++) {
		if (!str[i])
			return i + 1;
		if (str[i] == modetab[i].c)
			mode |= modetab[i].mask;
		else if (str[i] != '-')
			return i + 1;
	}
	if (str[i])
		return i + 1;
	*pmode = mode;
	return 0;
}

/* #pragma dbprop <name> [null] [mode] */
MF_PRAGMA(dbprop, 3, 4)
{
	int null = 0;
	mode_t mode = DEFAULT_DB_MODE;
	struct db_prop *prop;
	int rc;
	char *pat;

	--argc;
	pat = *++argv;
	
	while (--argc) {
		char *p = *++argv;

		if (strcmp(p, "null") == 0)
			null = 1;
		else if (c_isdigit(*p)) {
			unsigned long n = strtoul(p, &p, 8);
			if (*p || (mode = n) != n) {
				parse_error(_("bad numeric file mode"));
				return;
			}
		} else if (rc = strtomode(p, &mode)) {
			parse_error(_("bad symbolic file mode (near %s)"),
				    p + rc - 1);
			return;
		}
	}

	if (!db_prop_list) {
		rc = mu_list_create(&db_prop_list);
		if (rc) {
			parse_error(_("cannot create list: %s"),
				    mu_strerror(rc));
			return;
		}
		mu_list_set_comparator(db_prop_list, db_prop_compare);
	}

	if (mu_list_locate(db_prop_list, pat, (void**) &prop)) {
		prop = xmalloc(sizeof(*prop));
		prop->pat = xstrdup(pat);
		rc = mu_list_append(db_prop_list, prop);
		if (rc) {
			parse_error(_("Cannot create list: %s"),
				    mu_strerror(rc));
			return;
		}
	}
	prop->mode = mode;
	prop->null = null;
}

const struct db_prop *
db_prop_lookup(const char *name)
{
	mu_iterator_t itr;
	const struct db_prop *found = NULL;
	
	if (db_prop_list
	    && mu_list_get_iterator(db_prop_list, &itr) == 0) {
		for (mu_iterator_first(itr);
		     !mu_iterator_is_done(itr);
		     mu_iterator_next(itr)) {
			const struct db_prop *p;
			mu_iterator_current(itr, (void**)&p);
			if (strcmp(p->pat, name) == 0) {
				found = p;
				break;
			} else if (fnmatch(p->pat, name, 0) == 0)
				found = p;
		}
		mu_iterator_destroy(&itr);
	}
	return found;	
}


#define LOOKUP_NULL_BYTE 0x1
#define LOOKUP_TEST_ONLY 0x2

MF_DSEXP_SUPPRESS([<dbmap_lookup>],[<
static int
dbmap_lookup(eval_environ_t env, char *dbname, const char *keystr,
	     const char *defval, int flags)
{
	int rc;
	MF_DBM_FILE db;
	MF_DBM_DATUM key;
	MF_DBM_DATUM contents;

	if (!defval)
		defval = "";
	if (mf_dbm_open(dbname, &db, MU_STREAM_READ, 0, NULL))
		MF_THROW(mfe_dbfailure,
			 _("mf_dbm_open(%s) failed: %s"),
			 dbname,
			 mf_dbm_strerror());

	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MF_DATUM_PTR(key) = (void*) keystr;
	MF_DATUM_SIZE(key) = strlen(keystr);
	if (flags & LOOKUP_NULL_BYTE)
		MF_DATUM_SIZE(key)++;
	rc = mf_dbm_fetch(&db, key, &contents) == 0;
	MF_DEBUG(MU_DEBUG_TRACE5,
                 ("Looking up %s: %s", keystr, rc ? "true" : "false"));
	if (flags & LOOKUP_TEST_ONLY) 
		push(env, (STKVAL)rc);
	else {
		if (!rc) {
			if (defval)
				pushs(env, (STKVAL)defval);
			else
				push(env, 0);
		} else if (((char*)MF_DATUM_PTR(contents))[MF_DATUM_SIZE(contents)-1]) {
			size_t off;
			size_t len = MF_DATUM_SIZE(contents);
			char *s = MF_ALLOC_HEAP(off, len + 1);
			memcpy(s, MF_DATUM_PTR(contents), len);
			s[len] = 0;
			push(env, (STKVAL) off);
		} else
			pushs(env, MF_DATUM_PTR(contents));
	}
	mf_dbm_datum_free(&contents);
	mf_dbm_close(&db);
	return rc;
}
>])

MF_DSEXP
MF_DEFUN(dbmap, NUMBER, STRING dbname, STRING key, OPTIONAL, NUMBER null)
{
	const struct db_prop *prop = db_prop_lookup(dbname);
	dbmap_lookup(env, dbname, key, NULL,
		     LOOKUP_TEST_ONLY | 
		         (MF_OPTVAL(null, prop && prop->null)
			   ? LOOKUP_NULL_BYTE : 0));
}
END

MF_DEFUN(dbget, STRING, STRING dbname, STRING key, OPTIONAL,
	 STRING defval, NUMBER null)
{
	const struct db_prop *prop = db_prop_lookup(dbname);
	dbmap_lookup(env, dbname, key, 
                     MF_OPTVAL(defval), 
		     MF_OPTVAL(null, prop && prop->null)
		      ? LOOKUP_NULL_BYTE : 0);
}
END

MF_DEFUN(dbput, VOID, STRING dbname, STRING keystr, STRING value,
	 OPTIONAL, NUMBER null, NUMBER mode)
{
	int rc;
	MF_DBM_FILE db;
	MF_DBM_DATUM key;
	MF_DBM_DATUM contents;
	const struct db_prop *prop = db_prop_lookup(dbname);
	
	if (mf_dbm_open(dbname, &db, MU_STREAM_RDWR,
			MF_OPTVAL(mode, (prop ? prop->mode : DEFAULT_DB_MODE)),
			NULL))
		MF_THROW(mfe_dbfailure,
			 _("mf_dbm_open(%s) failed: %s"),
			 dbname,
			 mf_dbm_strerror());
	memset(&key, 0, sizeof key);
	MF_DATUM_PTR(key) = keystr;
	MF_DATUM_SIZE(key) = strlen(keystr);
	if (MF_OPTVAL(null, prop && prop->null))
		MF_DATUM_SIZE(key)++;

	memset(&contents, 0, sizeof contents);
	MF_DATUM_PTR(contents) = value;
	MF_DATUM_SIZE(contents) = strlen(value) + 1;
	
	rc = mf_dbm_insert(&db, key, contents, 1);
	mf_dbm_close(&db);
	MF_ASSERT(rc == 0,
		  mfe_dbfailure,
		  _("failed to insert data to %s: %s %s: %s"),
		  dbname,
		  keystr,
		  value,
		  mf_dbm_strerror());
}
END

MF_DEFUN(dbinsert, VOID, STRING dbname, STRING keystr, STRING value,
	 OPTIONAL, NUMBER replace, NUMBER null, NUMBER mode)
{
	int rc;
	MF_DBM_FILE db;
	MF_DBM_DATUM key;
	MF_DBM_DATUM contents;
	const struct db_prop *prop = db_prop_lookup(dbname);
	
	if (mf_dbm_open(dbname, &db, MU_STREAM_RDWR,
			MF_OPTVAL(mode, (prop ? prop->mode : DEFAULT_DB_MODE)),
			NULL))
		MF_THROW(mfe_dbfailure,
			 _("mf_dbm_open(%s) failed: %s"),
			 dbname,
			 mf_dbm_strerror());
	memset(&key, 0, sizeof key);
	MF_DATUM_PTR(key) = keystr;
	MF_DATUM_SIZE(key) = strlen(keystr);
	if (MF_OPTVAL(null, prop && prop->null))
		MF_DATUM_SIZE(key)++;

	memset(&contents, 0, sizeof contents);
	MF_DATUM_PTR(contents) = value;
	MF_DATUM_SIZE(contents) = strlen(value) + 1;
	
	rc = mf_dbm_insert(&db, key, contents, MF_OPTVAL(replace));
	mf_dbm_close(&db);
	if (rc == MU_ERR_EXISTS)
		MF_THROW(mfe_exists, _("record already exists"));
	
	MF_ASSERT(rc == 0,
		  mfe_dbfailure,
		  _("failed to insert data to %s: %s %s: %s"),
		  dbname,
		  keystr,
		  value,
		  mf_dbm_strerror());
}
END

MF_DEFUN(dbdel, VOID, STRING dbname, STRING keystr, OPTIONAL, NUMBER null,
	 NUMBER mode)
{
	MF_DBM_FILE db;
	MF_DBM_DATUM key;
	int rc;
	const struct db_prop *prop = db_prop_lookup(dbname);
	
	if (mf_dbm_open(dbname, &db, MU_STREAM_RDWR,
			MF_OPTVAL(mode, (prop ? prop->mode : DEFAULT_DB_MODE)),
			NULL))
		MF_THROW(mfe_dbfailure,
			 _("mf_dbm_open(%s) failed: %s"),
			 dbname,
			 mf_dbm_strerror());
	memset(&key, 0, sizeof key);
	MF_DATUM_PTR(key) = keystr;
	MF_DATUM_SIZE(key) = strlen(keystr);
	if (MF_OPTVAL(null, prop && prop->null))
		MF_DATUM_SIZE(key)++;
	rc = mf_dbm_delete(&db, key);
	mf_dbm_close(&db);
	MF_ASSERT(rc == 0,
                  mfe_dbfailure,
                  _("failed to delete data `%s' from `%s': %s"),
		  keystr,
                  dbname,
		  mf_dbm_strerror());
}
END


#define NUMDB 128

struct db_tab {
	int used;
	MF_DBM_FILE db;
	MF_DBM_DATUM key;
};

static void *
alloc_db_tab()
{
	return xcalloc(NUMDB, sizeof(struct db_tab));
}

static void
close_db_tab(struct db_tab *dbt)
{
	if (dbt->used) {
		mf_dbm_datum_free(&dbt->key);
		mf_dbm_close(&dbt->db);
		dbt->used = 0;
	}
}

static void
destroy_db_tab(void *data)
{
	int i;
	struct db_tab *db = data;
	for (i = 0; i < NUMDB; i++) 
		close_db_tab(db + i);
	free(db);
}

MF_DECLARE_DATA(DBTAB, alloc_db_tab, destroy_db_tab);

static int
new_db_tab(struct db_tab *dbt)
{
	int i;
	for (i = 0; i < NUMDB; i++)
		if (!dbt[i].used) {
			dbt[i].used = 1;
			return i;
		}
	return -1;
}
			


MF_DEFUN(dbfirst, NUMBER, STRING dbname)
{
	int rc;
	int n;
	struct db_tab *dbt = MF_GET_DATA;
	MF_DBM_FILE db;
	MF_DBM_DATUM key;
	
	if (mf_dbm_open(dbname, &db, MU_STREAM_READ, 0, NULL))
		MF_THROW(mfe_dbfailure,
			 _("mf_dbm_open(%s) failed: %s"),
			 dbname,
			 mf_dbm_strerror());
	rc = mf_dbm_firstkey(&db, &key);
	MF_ASSERT(rc == 0, mfe_dbfailure,
		  _("mf_dbm_firstkey failed: %s"),
		  mf_dbm_strerror());
	n = new_db_tab(dbt);
	MF_ASSERT(n >= 0,
		  mfe_failure,
		 _("no more database entries available"));
	dbt += n;
	dbt->db = db;
	dbt->key = key;
	MF_RETURN(n);
}
END

MF_DEFUN(dbnext, NUMBER, NUMBER dn)
{
	struct db_tab *dbt = MF_GET_DATA + dn;
	MF_DBM_DATUM nextkey;
	int rc;
	
	MF_ASSERT(dn >= 0 && dn < NUMDB && dbt->used,
		  mfe_range,
		  _("invalid database descriptor"));

	rc = mf_dbm_nextkey(&dbt->db, dbt->key, &nextkey);
	if (rc) {
		close_db_tab(dbt);
		MF_RETURN(0);
	}
	mf_dbm_datum_free(&dbt->key);
	dbt->key = nextkey;
	MF_RETURN(1);
}
END

MF_DEFUN(dbkey, STRING, NUMBER dn)
{
	size_t off, len;
	char *s;
	struct db_tab *dbt = MF_GET_DATA + dn;
	
	MF_ASSERT(dn >= 0 && dn < NUMDB && dbt->used,
		  mfe_range,
		  _("invalid database descriptor"));
	
	len = MF_DATUM_SIZE(dbt->key);
	s = MF_ALLOC_HEAP(off, len + 1);
	memcpy(s, MF_DATUM_PTR(dbt->key), len);
	s[len] = 0;
	MF_RETURN(off);
}
END

MF_DEFUN(dbvalue, STRING, NUMBER dn)
{
	int rc;
	size_t off, len;
	char *s;
	struct db_tab *dbt = MF_GET_DATA + dn;
	MF_DBM_DATUM contents;
	
	MF_ASSERT(dn >= 0 && dn < NUMDB && dbt->used,
		  mfe_range,
		  _("invalid database descriptor"));

	memset(&contents, 0, sizeof contents);
	rc = mf_dbm_fetch(&dbt->db, dbt->key, &contents);
	MF_ASSERT(rc == 0,
		  mfe_dbfailure,
		  _("key not found: %s"),
		  mf_dbm_strerror());
	
	len = MF_DATUM_SIZE(contents);
	s = MF_ALLOC_HEAP(off, len + 1);
	memcpy(s, MF_DATUM_PTR(contents), len);
	s[len] = 0;
	mf_dbm_datum_free(&contents);
	MF_RETURN(off);
}
END


enum greylist_semantics
{
	greylist_traditional,
	greylist_ct
};

static enum greylist_semantics greylist_semantics = greylist_traditional;

/* #pragma greylist {traditional|gray|ct|con-tassios}*/
MF_PRAGMA(greylist, 2, 2)
{
	if (strcmp(argv[1], "traditional") == 0
	    || strcmp(argv[1], "gray") == 0)
		greylist_semantics = greylist_traditional;
	else if (strcmp(argv[1], "ct") == 0
		 || strcmp(argv[1], "con-tassios") == 0)
		greylist_semantics = greylist_ct;
	else
		/* TRANSLATORS: Do not translate keywords:
		   traditional, gray, ct, con-tassios */
		parse_error(_("unknown semantics; allowed values are: "
			      "traditional (or gray) and "
			      "ct (or con-tassios)"));
}

/* FIXME: Duplicated in lib/cache.c */
static char *
format_timestr(time_t timestamp, char *timebuf, size_t bufsize)
{
	struct tm tm;
	gmtime_r(&timestamp, &tm);
	strftime(timebuf, bufsize, "%c", &tm);
	return timebuf;
}

MF_VAR(greylist_seconds_left, NUMBER);

/* The traditional (aka gray's) greylist implementation: the greylist
   database keeps the time the greylisting was activated.
 */
static int
do_greylist_traditional(eval_environ_t env, char *email, long interval)
{
	int rc;
	MF_DBM_FILE db;
	MF_DBM_DATUM key;
	MF_DBM_DATUM contents;
	int readonly;
	time_t now;

	rc = mf_dbm_open(greylist_format->dbname, &db, MU_STREAM_RDWR, 0600,
			 &readonly);
        MF_ASSERT(rc == 0, mfe_dbfailure, _("mf_dbm_open(%s) failed: %s"),
			      greylist_format->dbname, mf_dbm_strerror());
	
	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MF_DATUM_PTR(key) = email;
	MF_DATUM_SIZE(key) = strlen(email)+1;

	time(&now);
	if (mf_dbm_fetch(&db, key, &contents) == 0) {
		time_t timestamp, diff;

		MF_ASSERT(MF_DATUM_SIZE(contents) == sizeof timestamp,
			 mfe_dbfailure,
			  _("greylist database %s has wrong data size"),
			 greylist_format->dbname);
		
		timestamp = *(time_t*) MF_DATUM_PTR(contents);
		diff = now - timestamp;

		if (mu_debug_level_p(debug_handle, MU_DEBUG_TRACE5)) {
			char timebuf[32];
			mu_debug_log("%s entered greylist database on %s, "
				     "%ld seconds ago",
				     email,
				     format_timestr(timestamp, timebuf,
						    sizeof timebuf),
				     (long) diff);
		}

		if (diff < interval) {
			diff = interval - diff;

			MF_VAR_REF(greylist_seconds_left, diff);
			
			MF_DEBUG(MU_DEBUG_TRACE6, 
                                 ("%s still greylisted (for %lu sec.)",
			          email,
			          (unsigned long) diff));
			rc = 1;
		} else if (diff > greylist_format->expire_interval) {
			MF_DEBUG(MU_DEBUG_TRACE6, 
                                 ("greylist record for %s expired", email));
			MF_VAR_REF(greylist_seconds_left, interval);
			if (!readonly) {
				memcpy(MF_DATUM_PTR(contents),
				       &now, sizeof now);
				if (mf_dbm_insert(&db, key, contents, 1))
					mu_error(_("cannot insert datum `%-.*s' in "
						   "greylist database %s: %s"),
			                         MF_DATUM_SIZE(key),
						 (char*)MF_DATUM_PTR(key),
						 greylist_format->dbname,
	                                         mf_dbm_strerror());
			} else
				MF_DEBUG(MU_DEBUG_TRACE6, 
                                         ("database opened in readonly mode: "
			  	          "not updating"));
			rc = 1;
		} else {
			MF_DEBUG(MU_DEBUG_TRACE6, 
                                 ("%s finished greylisting period", email));
			rc = 0;
		}
		mf_dbm_datum_free(&contents);
	} else if (!readonly) {
		MF_DEBUG(MU_DEBUG_TRACE6, ("greylisting %s", email));
		MF_VAR_REF(greylist_seconds_left, interval);
		MF_DATUM_PTR(contents) = (void*)&now;
		MF_DATUM_SIZE(contents) = sizeof now;
		if (mf_dbm_insert(&db, key, contents, 1))
			mu_error(_("Cannot insert datum `%-.*s' in greylist "
				   "database %s: %s"),
		                 MF_DATUM_SIZE(key), (char*)MF_DATUM_PTR(key),
				 greylist_format->dbname,
	                         mf_dbm_strerror());
		rc = 1;
	} else
		rc = 0;

	mf_dbm_close(&db);

	return rc;
}

/* Implementation of the is_greylisted predicate has no sense for 
   traditional greylist databases, because greylisting interval is
   not known beforehand.
   FIXME: keep the reference below up to date.
 */
static int
is_greylisted_traditional(eval_environ_t env, char *email)
{
	MF_THROW(mfe_failure,
		 _("is_greylisted is not implemented for traditional greylist databases; "
		   "see documentation, chapter %s %s for more info"),
		 "4.12.1.23", "Greylisting functions");
	return 0;
}

/* New greylist implementation (by Con Tassios): the database keeps
   the time the greylisting period is set to expire (`interval' seconds
   from now)
*/
static int
do_greylist_ct(eval_environ_t env, char *email, long interval)
{
	int rc;
	MF_DBM_FILE db;
	MF_DBM_DATUM key;
	MF_DBM_DATUM contents;
	int readonly;
	time_t now;

	rc = mf_dbm_open(greylist_format->dbname, &db, MU_STREAM_RDWR, 0600,
			 &readonly);
        MF_ASSERT(rc == 0, mfe_dbfailure, _("mf_dbm_open(%s) failed: %s"),
			      greylist_format->dbname, mf_dbm_strerror());
	
	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MF_DATUM_PTR(key) = email;
	MF_DATUM_SIZE(key) = strlen(email)+1;

	time(&now);
	if (mf_dbm_fetch(&db, key, &contents) == 0) {
		time_t timestamp;

		MF_ASSERT(MF_DATUM_SIZE(contents) == sizeof timestamp,
			 mfe_dbfailure,
			  _("greylist database %s has wrong data size"),
			 greylist_format->dbname);
		
		timestamp = *(time_t*) MF_DATUM_PTR(contents);

		if (now < timestamp) {
			time_t diff = timestamp - now;
			MF_VAR_REF(greylist_seconds_left, diff);
			
			MF_DEBUG(MU_DEBUG_TRACE6, 
                                 ("%s still greylisted (for %lu sec.)",
			          email,
                                  (unsigned long) diff));
			rc = 1;
		} else if (now - timestamp >
			   greylist_format->expire_interval) {
			MF_DEBUG(MU_DEBUG_TRACE6, 
                                 ("greylist record for %s expired", email));
			MF_VAR_REF(greylist_seconds_left, interval);
			if (!readonly) {
				now += interval;
				memcpy(MF_DATUM_PTR(contents),
				       &now, sizeof now);
				if (mf_dbm_insert(&db, key, contents, 1))
					mu_error(_("Cannot insert datum "
						   "`%-.*s' in greylist "
						   "database %s: %s"),
			                         MF_DATUM_SIZE(key),
						 (char*)MF_DATUM_PTR(key),
						 greylist_format->dbname,
	                                         mf_dbm_strerror());
			} else
				MF_DEBUG(MU_DEBUG_TRACE6, 
                                         ("database opened in readonly mode: "
				          "not updating"));
			rc = 1;
		} else {
			MF_DEBUG(MU_DEBUG_TRACE6,
				 ("%s finished greylisting period", email));
			rc = 0;
		}
		mf_dbm_datum_free(&contents);
	} else if (!readonly) {
		MF_DEBUG(MU_DEBUG_TRACE6, ("greylisting %s", email));
		MF_VAR_REF(greylist_seconds_left, interval);
		now += interval;
		MF_DATUM_PTR(contents) = (void*)&now;
		MF_DATUM_SIZE(contents) = sizeof now;
		if (mf_dbm_insert(&db, key, contents, 1))
			mu_error(_("Cannot insert datum `%-.*s' in greylist "
				   "database %s: %s"),
		                 MF_DATUM_SIZE(key), (char*)MF_DATUM_PTR(key),
				 greylist_format->dbname,
	                         mf_dbm_strerror());
		rc = 1;
	} else
		rc = 0;
	
	mf_dbm_close(&db);
	
	return rc;
}

/* The `is_greylisted' predicate for new databases */
static int
is_greylisted_ct(eval_environ_t env, char *email)
{
	int rc;
	MF_DBM_FILE db;
	MF_DBM_DATUM key;
	MF_DBM_DATUM contents;
	int readonly;
	time_t now;

	rc = mf_dbm_open(greylist_format->dbname, &db, MU_STREAM_RDWR, 0600,
			 &readonly);
	MF_ASSERT(rc == 0, mfe_dbfailure, _("mf_dbm_open(%s) failed: %s"),
		  greylist_format->dbname, mf_dbm_strerror());

	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MF_DATUM_PTR(key) = email;
	MF_DATUM_SIZE(key) = strlen(email) + 1;

	time(&now);
	if (mf_dbm_fetch(&db, key, &contents) == 0) {
		time_t timestamp;

		MF_ASSERT(MF_DATUM_SIZE(contents) == sizeof timestamp,
			mfe_dbfailure,
			  _("greylist database %s has wrong data size"),
			greylist_format->dbname);
                
		timestamp = *(time_t*) MF_DATUM_PTR(contents);
                
		rc = timestamp > now;
                if (rc)
			MF_VAR_REF(greylist_seconds_left, timestamp - now);

		mf_dbm_datum_free(&contents);
	} else  
		rc = 0;

	mf_dbm_close(&db);

	return rc;
}

struct greylist_class {
	int (*gl_fun)(eval_environ_t, char *, long);
	int (*gl_pred)(eval_environ_t, char *);
};

struct greylist_class greylist_class[] = {
	{ do_greylist_traditional, is_greylisted_traditional },
	{ do_greylist_ct, is_greylisted_ct }
};

/* greylist(key, interval)

   Returns true if the key is greylisted, false if it's OK to
   deliver mail.
 */
MF_DEFUN(greylist, NUMBER, STRING email, NUMBER interval)
{
	MF_RETURN(greylist_class[greylist_semantics].gl_fun(env, email,
							    interval));
}
END

/* is_greylisted(key)
   
   Returns true if the key is greylisted, otherwise false
   
*/
MF_DEFUN(is_greylisted, NUMBER, STRING email)
{
	MF_RETURN(greylist_class[greylist_semantics].gl_pred(env, email));
}
END

MF_DEFUN(db_name, STRING, STRING fmtid)
{
	struct db_format *fmt = db_format_lookup(fmtid);
	MF_ASSERT(fmt != NULL,
		  mfe_not_found,
		  _("no such db format: %s"), fmtid);
	MF_RETURN_STRING(fmt->dbname);
}
END

MF_DEFUN(db_get_active, NUMBER, STRING fmtid)
{
	struct db_format *fmt = db_format_lookup(fmtid);
	MF_ASSERT(fmt != NULL,
		  mfe_not_found,
		  _("no such db format: %s"), fmtid);
	MF_RETURN(fmt->enabled);
}
END

MF_DEFUN(db_set_active, VOID, STRING fmtid, NUMBER active)
{
	struct db_format *fmt = db_format_lookup(fmtid);
	MF_ASSERT(fmt != NULL,
		  mfe_not_found,
		  _("no such db format: %s"), fmtid);
	fmt->enabled = active;
}
END

MF_DEFUN(db_expire_interval, NUMBER, STRING fmtid)
{
	struct db_format *fmt = db_format_lookup(fmtid);
	MF_ASSERT(fmt != NULL,
		  mfe_not_found,
		  _("no such db format: %s"), fmtid);
	MF_RETURN(fmt->expire_interval);
}
END

MF_INIT

